/*
 *	Ohio LAM
 *
 *	Function:	- master program of mandelbrot demo
 *			- distributes work orders to the slaves
 *			- collects pixel buffers from the slaves
 *			- sends results to xmtv(1)
 */

#include <errno.h>
#include <stdlib.h>

#include <mpi.h>
#include <tvreq.h>
#include "mandel.h"

/*
 * local functions
 */
static int		harvest();		/* gets results */
static void		seed();			/* distributes work */

/*
 * local variables
 */
static int		grid_dim[2];		/* grid size in pixels */
static int		region_dim[2];		/* region size in pixels */
static char		*region;		/* color region buffer */
static int		myrank;			/* compute task's rank */


main(argc, argv)

int			argc;
char			*argv[];

{
	float		grid_coord[4];		/* complex grid description */
	float		init[4];		/* scale and translate data */
	int		n;
	int		ntasks;			/* # compute tasks */
	int		rank;			/* compute task's rank */
	int		x, y;			/* pixel coordinates */
	struct tvstate	tvinfo;			/* screen information */
/*
 * Initialize MPI.
 */
	MPI_Init(&argc, &argv);
	MPI_Comm_size(MPI_COMM_WORLD, &ntasks);
	MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
/*
 * Get frame buffer information and clear the screen.
 */
	if (lam_tvstate(&tvinfo)) MPI_Abort(MPI_COMM_WORLD, errno);

	if (lam_tvclr()) MPI_Abort(MPI_COMM_WORLD, errno);
/*
 * Hardcode the universe.
 */
	grid_coord[0] = -0.74;		/* up-left real */
	grid_coord[1] = 0.25;		/* up-left imag. */
	grid_coord[2] = -0.71;		/* down-right real */
	grid_coord[3] = 0.20;		/* down-right imag. */

	grid_dim[0] = tvinfo.tvs_xlen;
	grid_dim[1] = tvinfo.tvs_ylen;

	region_dim[0] = 20;
	region_dim[1] = 20;
/*
 * Allocate space for region.
 */
	region = malloc((unsigned) (region_dim[0] * region_dim[1]));
	if (region == 0) MPI_Abort(MPI_COMM_WORLD, errno);
/*
 * Send translate and scale factors to each slave.
 */
	init[0] = (grid_coord[2] - grid_coord[0]) / grid_dim[0];
	init[1] = -((grid_coord[1] - grid_coord[3]) / grid_dim[1]);
	init[2] = grid_coord[0];
	init[3] = grid_coord[1];

	for (rank = 0; rank < ntasks; rank++) {

		if (rank != myrank) {
			MPI_Send(init, 4, MPI_FLOAT, rank,
					WORKTAG, MPI_COMM_WORLD);
		}
	}
/*
 * Distribute first work order to all slaves.
 */
	n = 0;
	y = 0;

	while ((y < grid_dim[1]) && (n < ntasks)) {

		for (x = 0; (x < grid_dim[0]) && (n < ntasks); ) {

			if (n == myrank) {
				++n;
			}

			if (n < ntasks) {
				seed(n++, x, y);
				x += region_dim[0];
			}
		}

		if (n < ntasks) {
			y += region_dim[1];
		}
	}
/*
 * Collect a result and return a new work order until the
 * entire grid is computed.
 */
	for (; y < grid_dim[1]; y += region_dim[1]) {

		for (; x < grid_dim[0]; x += region_dim[0]) {
			n = harvest();
			seed(n, x, y);
		}

		x = 0;
	}
/*
 * Collect remaining results.
 */
	for (n = 0; n < ntasks - 1; ++n) {
		harvest();
	}
/*
 * We have all the answers now, so kill the workers.
 */
	for (rank = 0; rank < ntasks; rank++) {

		if (rank != myrank) {
			MPI_Send(0, 0, MPI_INT, rank, DIETAG, MPI_COMM_WORLD);
		}
	}

	MPI_Finalize();
	exit(0);
}

/*
 *	harvest
 *
 *	Function:	- receives results from one slave
 *	Returns:	- slave's rank
 */
static int
harvest()

{
	int		len0, len1;		/* x,y dim region length */
	int		i;
	MPI_Status	stat;			/* status returned by recv */
	int		source;			/* who sent last message */
	char		*pgrid;			/* ptr within grid buffer */
	char		*pregion;		/* ptr within region buffer */
	int		result_dim[4];		/* result dimensions */

	MPI_Recv(result_dim, 4, MPI_INT, MPI_ANY_SOURCE,
			MPI_ANY_TAG, MPI_COMM_WORLD, &stat);

	source = stat.MPI_SOURCE;

	MPI_Recv(region, region_dim[0] * region_dim[1], MPI_CHAR,
			source, MPI_ANY_TAG, MPI_COMM_WORLD, &stat);
		
	len0 = result_dim[2] - result_dim[0] + 1;
	len1 = result_dim[3] - result_dim[1] + 1;

	if (lam_tvrect(result_dim[0], result_dim[1], len0, len1, region))
			MPI_Abort(MPI_COMM_WORLD, errno);

	return(source);
}

/*
 *	seed
 *
 *	Function:	- sends out work request
 */
static void
seed(rank, x, y)

int			rank;
int			x, y;

{
	int		work[4];	/* network message desc. */

	work[0] = x;			/* pixel coord's */
	work[1] = y;
	work[2] = (x + region_dim[0] > grid_dim[0]) ?
			grid_dim[0] - 1 : x + region_dim[0] - 1;
	work[3] = (y + region_dim[1] > grid_dim[1]) ?
			grid_dim[1] - 1 : y + region_dim[1] - 1;

	MPI_Send(work, 4, MPI_INT, rank, WORKTAG, MPI_COMM_WORLD);
}
